import ROOT
import os
import numpy as np
import pandas as pd
import argparse
import json
from matplotlib import pyplot as plt
from ROOT import gROOT, gSystem, gInterpreter, gStyle, gPad, kRed, kAzure
from plot_utils import prepare_canvas, prepare_figure, my_style, get_hist_contents


# make sure the relative path for root scripts is correct
owd = os.getcwd()
script_dir = os.path.dirname(os.path.realpath(__file__))
os.chdir(os.path.join(script_dir, '..'))
print('working directory is ' + os.getcwd())

# rootlogon macro
gROOT.Macro('rootlogon.C')

# the functions that process data are defined in scripts/root/dataframe_analysis.C
func_code = open(os.path.join(script_dir, 'root', 'dataframe_analysis.C')).read()
gInterpreter.Declare(func_code)

# global style
my_style.cd()
bprops = dict(boxstyle='round', facecolor='wheat', alpha=0.2)

# argument parser
parser = argparse.ArgumentParser('timing cuts study')
parser.add_argument('root_file', help='root file output from analyzer')
parser.add_argument('json_file', help='output timing cut in json')
parser.add_argument('-b', dest='batch', action='store_true', help='set to batch mode')
parser.add_argument('--read-cuts', dest='read', action='store_true', help='directly use cuts defined in json file.')
parser.add_argument('--snapshot', dest='snap', default='', type=str, help='path to save the data with trigger cuts')
parser.add_argument('--plot-dir', dest='out_dir', default='.', type=str, help='directory to save the plot')
parser.add_argument('--layout', dest='layout', default=os.path.join(script_dir, '..', 'database/channels_layout.json'),
                    help='json data for channels layout')
args = parser.parse_args()

# recover paths
for attr in ['root_file', 'json_file', 'snap', 'out_dir', 'layout']:
    setattr(args, attr, os.path.join(owd, getattr(args, attr)))

os.makedirs(args.out_dir, exist_ok=True)

# batch mode
gROOT.SetBatch(args.batch)

# root dataframe
rdf = ROOT.RDataFrame('EvTree', args.root_file)
save_list = [
    'ecal', 'ecal_ch',
    # 'scint2', 'scint2_ch'
]

# read json
json_data = dict()
if os.path.exists(args.json_file):
    with open(args.json_file, 'r') as f:
        json_data.update(json.load(f))

with open(args.layout, 'r') as f:
    channels_layout = json.load(f)


# main function
def timing_cuts_search(df, cuts_dict, cut_width, hbins=100):
    for ch, cuts in cuts_dict.items():
        search = cuts['cut']
        ref = cuts['ref']
        hist = df.Define('peak_time', 'unpack_time({}, true, {})'.format(ch, ref))\
                 .Histo1D(ROOT.RDF.TH1DModel('time_hist1', '', hbins, *search), 'peak_time')

        pos = hist.GetBinCenter(hist.GetMaximumBin())
        height = hist.GetMaximum()
        if height > 0:
            cuts_dict[ch]['cut'] = (pos - cut_width, pos + cut_width)
    return cuts_dict


# plot distributions with cuts
def plot_dist_tcuts(df, canvas, pads_dict, cuts_dict, attr, hrange=(128, -256, 256), use_ref=False, hthres=0):
    stacks, hists = [], []
    for ch, pad in pads_dict.items():
        cut = cuts_dict[ch]['cut']
        ref = cuts_dict[ch]['ref'] if use_ref else '0'
        stacks.append(ROOT.THStack('hs_' + ch, ch))
        attr_name = '{}_{}'.format(ch, attr)
        hists.append(
            df.Define(attr_name, 'unpack_{}({}, true, {})'.format(attr, ch, ref))\
              .Histo1D(ROOT.RDF.TH1DModel(attr_name + '_h', '', *hrange), attr_name)
        )
        hists[-1].SetFillColorAlpha(38, 0.7)
        hists[-1].SetFillStyle(3000)
        hists[-1].SetLineStyle(2)
        hists[-1].SetLineColor(kAzure - 6)
        stacks[-1].Add(hists[-1].GetPtr())

        # cut peaks
        hists.append(
            df.Filter('{}_peak.height > {}'.format(ch, hthres))\
              .Define(attr_name + '_cut', '{}_peak.{} - {}'.format(ch, attr, ref))\
              .Histo1D(ROOT.RDF.TH1DModel(attr_name + '_hcut', '', *hrange), attr_name + '_cut')
        )
        hists[-1].SetFillColorAlpha(46, 0.7)
        hists[-1].SetFillStyle(3000)
        hists[-1].SetLineStyle(2)
        hists[-1].SetLineColor(kRed + 3)
        stacks[-1].Add(hists[-1].GetPtr())

        pad.cd()
        stacks[-1].Draw('nostack')
        canvas.Update()
    return stacks, hists



# =============================================================================
# MaPMT sum channel timing study
# =============================================================================
print('Study timing cuts for MaPMT sum channels...')
# calorimeter (trigger) channels
sum_channels = channels_layout['MaPMT Sums']
# peak search range
if args.read:
    sum_cuts = {c: json_data[c] for chs in sum_channels for c in chs}
else:
    sum_search = {c: {'cut': (-100, -50), 'ref': 'ecal.time'} for chs in sum_channels for c in chs}
    # change sum timing search range for special cases
    sum_search['Cer43_5']['cut'] = (-150, -100)
    sum_cuts = timing_cuts_search(rdf, sum_search, cut_width=8)
    json_data.update(sum_cuts)

sum_height_cut = 0

# apply cuts
for ch, cut in sum_cuts.items():
    rdf = rdf.Define('{}_peak'.format(ch), 'get_maximum_peak({}.peaks, time_cut_rel, {}, {}, {})'\
                                           .format(ch, *cut['cut'], cut['ref']))
    save_list.append('{}_peak'.format(ch))

# plot timing distributions with t-cuts
c1 = ROOT.TCanvas('sum_time', '', 1920, 1080)
pads1 = prepare_canvas(c1, sum_channels, pad_label='MaPMT Sum Peak Time',
                       x_label='Time Rel. to ECal. (ns)', y_label='Counts')
res1 = plot_dist_tcuts(rdf, c1, pads1, sum_cuts, 'time', hrange=(80, -160, 160), hthres=sum_height_cut, use_ref=True)
c1.SaveAs(os.path.join(args.out_dir, 'tcuts_sum_time.png'))

# plot height distributions with t-cuts
c2 = ROOT.TCanvas('sum_height', '', 1920, 1080)
pads2 = prepare_canvas(c2, sum_channels, pad_label='MaPMT Sum Peak Height', x_label='ADC Value', y_label='Counts')
res2 = plot_dist_tcuts(rdf, c2, pads2, sum_cuts, 'height', hrange=(100, 0, 1000), hthres=sum_height_cut)
c2.SaveAs(os.path.join(args.out_dir, 'tcuts_sum_height.png'))



# =============================================================================
# MaPMT quadrant channel timing study
# =============================================================================
print('Study timing cuts for MaPMT quadrant channels...')
# calorimeter (trigger) channels
quad_channels = channels_layout['MaPMT Quads']
# peak search range
if args.read:
    quad_cuts = {c: json_data[c] for chs in quad_channels for c in chs}
else:
    quad_search = {c:
        {'cut': (-80, 20), 'ref': c.split('_')[0] + '_5_peak.time', 'group': c.split('_')[0]}
        for chs in quad_channels for c in chs}
    quad_cuts = timing_cuts_search(rdf, quad_search, cut_width=5)
    json_data.update(quad_cuts)

quad_height_cut = 0

# apply cuts
for ch, cut in quad_cuts.items():
    rdf = rdf.Define('{}_peak'.format(ch), 'get_maximum_peak({}.peaks, time_cut_rel, {}, {}, {})'\
                                           .format(ch, *cut['cut'], cut['ref']))
    save_list.append('{}_peak'.format(ch))

# plot timing distributions with t-cuts
c3 = ROOT.TCanvas('quad_time', '', 1920, 1080)
pads3 = prepare_canvas(c3, quad_channels, pad_label='MaPMT Quad. Peak Time',
                       x_label='Time Rel. to Sum (ns)', y_label='Counts')
res3 = plot_dist_tcuts(rdf, c3, pads3, quad_cuts, 'time', hrange=(80, -160, 160), hthres=quad_height_cut, use_ref=True)
c3.SaveAs(os.path.join(args.out_dir, 'tcuts_quad_time.png'))

# plot height distributions with t-cuts
c4 = ROOT.TCanvas('quad_height', '', 1920, 1080)
pads4 = prepare_canvas(c4, quad_channels, pad_label='MaPMT Quad. Peak Height', x_label='ADC Value', y_label='Counts')
res4 = plot_dist_tcuts(rdf, c4, pads4, quad_cuts, 'height', hrange=(100, 0, 1000), hthres=quad_height_cut)
c4.SaveAs(os.path.join(args.out_dir, 'tcuts_quad_height.png'))



# =============================================================================
# Ending
# =============================================================================
rdf.Report().Print()
# save json
if not args.read:
    with open(args.json_file, 'w') as f:
        json.dump(json_data, f, indent=2)

# save data
if args.snap:
    print('Saving data snapshot to \"{}\", with columns:\n{}'.format(args.snap, save_list))
    rdf.Snapshot('EvTree', args.snap, save_list)

# this keeps the canvas
if not args.batch:
    print("press enter to quit")
    input()

