import ROOT
import os
import numpy as np
import pandas as pd
import argparse
import json
from matplotlib import pyplot as plt
from ROOT import gROOT, gSystem, gInterpreter, gStyle, gPad, kRed, kAzure
from plot_utils import prepare_canvas, prepare_figure, my_style, get_hist_contents


# make sure the relative path for root scripts is correct
owd = os.getcwd()
script_dir = os.path.dirname(os.path.realpath(__file__))
os.chdir(os.path.join(script_dir, '..'))
print('working directory is ' + os.getcwd())

# rootlogon macro
gROOT.Macro('rootlogon.C')

# the functions that process data are defined in scripts/root/dataframe_analysis.C
func_code = open(os.path.join(script_dir, 'root', 'dataframe_analysis.C')).read()
gInterpreter.Declare(func_code)

# global style
my_style.cd()
bprops = dict(boxstyle='round', facecolor='wheat', alpha=0.2)

# argument parser
parser = argparse.ArgumentParser('timing cuts study')
parser.add_argument('root_file', help='root file output from analyzer')
parser.add_argument('json_file', help='output timing cut in json')
parser.add_argument('-b', dest='batch', action='store_true', help='set to batch mode')
parser.add_argument('--read-cuts', dest='read', action='store_true', help='directly use cuts defined in json file.')
parser.add_argument('--snapshot', dest='snap', default='', type=str, help='path to save the data with trigger cuts')
parser.add_argument('--plot-dir', dest='out_dir', default='.', type=str, help='directory to save the plot')
parser.add_argument('--layout', dest='layout', default=os.path.join(script_dir, '..', 'database/channels_layout.json'),
                    help='json data for channels layout')
args = parser.parse_args()

# recover paths
for attr in ['root_file', 'json_file', 'snap', 'out_dir', 'layout']:
    setattr(args, attr, os.path.join(owd, getattr(args, attr)))

os.makedirs(args.out_dir, exist_ok=True)

# batch mode
gROOT.SetBatch(args.batch)

# root dataframe
rdf = ROOT.RDataFrame('EvTree', args.root_file)
# save_list = ['Cer{}{}_{}'.format(i, j, k) for i in np.arange(1, 5) for j in np.arange(1, 5) for k in np.arange(1, 6)]
save_list = ['LAPPD_{}{}'.format(l, i) for i in np.arange(1, 9) for l in 'ABCDEFGH']

# read json
json_data = dict()
if os.path.exists(args.json_file):
    with open(args.json_file, 'r') as f:
        json_data.update(json.load(f))

with open(args.layout, 'r') as f:
    channels_layout = json.load(f)

# main function
def timing_cuts_search(df, cuts_dict, cut_width, hbins=100):
    for ch, cuts in cuts_dict.items():
        search = cuts['cut']
        ref = cuts['ref']
        hist = df.Define('peak_time', 'unpack_time({}, true, {})'.format(ch, ref))\
                 .Histo1D(ROOT.RDF.TH1DModel('time_hist1', '', hbins, *search), 'peak_time')

        pos = hist.GetBinCenter(hist.GetMaximumBin())
        height = hist.GetMaximum()
        if height > 0:
            cuts_dict[ch]['cut'] = (pos - cut_width, pos + cut_width)
    return cuts_dict


# plot distributions with cuts
def plot_dist_tcuts(df, canvas, pads_dict, cuts_dict, attr, hrange=(128, -256, 256), use_ref=False, hthres=0):
    stacks, hists = [], []
    for ch, pad in pads_dict.items():
        cut = cuts_dict[ch]['cut']
        ref = cuts_dict[ch]['ref'] if use_ref else '0'
        stacks.append(ROOT.THStack('hs_' + ch, ch))
        attr_name = '{}_{}'.format(ch, attr)
        hists.append(
            df.Define(attr_name, 'unpack_{}({}, true, {})'.format(attr, ch, ref))\
              .Histo1D(ROOT.RDF.TH1DModel(attr_name + '_h', '', *hrange), attr_name)
        )
        hists[-1].SetFillColorAlpha(38, 0.7)
        hists[-1].SetFillStyle(3000)
        hists[-1].SetLineStyle(2)
        hists[-1].SetLineColor(kAzure - 6)
        stacks[-1].Add(hists[-1].GetPtr())

        # cut peaks
        hists.append(
            df.Filter('{}_peak.height > {}'.format(ch, hthres))\
              .Define(attr_name + '_cut', '{}_peak.{} - {}'.format(ch, attr, ref))\
              .Histo1D(ROOT.RDF.TH1DModel(attr_name + '_hcut', '', *hrange), attr_name + '_cut')
        )
        hists[-1].SetFillColorAlpha(46, 0.7)
        hists[-1].SetFillStyle(3000)
        hists[-1].SetLineStyle(2)
        hists[-1].SetLineColor(kRed + 3)
        stacks[-1].Add(hists[-1].GetPtr())

        pad.cd()
        stacks[-1].Draw('nostack')
        canvas.Update()
    return stacks, hists



# =============================================================================
# ECal channel timing study
# =============================================================================
print('Study timing cuts for ECal channels...')
# calorimeter (trigger) channels
ecal_channels = channels_layout['Calorimeter']
# peak search range
if args.read:
    ecal_cuts = {c: json_data[c] for chs in ecal_channels for c in chs}
else:
    ecal_search = {c: {'cut': (105, 135), 'ref': '0'} for chs in ecal_channels for c in chs}
    # change ecal timing search range for special cases
    # ecal_search['C4']['cut'] = (100, 130)
    ecal_cuts = timing_cuts_search(rdf, ecal_search, cut_width=10)
    json_data.update(ecal_cuts)

ecal_height_cut = 0

# apply cuts
for ch, cut in ecal_cuts.items():
    rdf = rdf.Define('{}_peak'.format(ch), 'get_maximum_peak({}.peaks, time_cut_rel, {}, {}, {})'\
                                           .format(ch, *cut['cut'], cut['ref']))

# plot timing distributions with t-cuts
c1 = ROOT.TCanvas('ECal_time', '', 1920, 1080)
pads1 = prepare_canvas(c1, ecal_channels, pad_label='Calorimeter Peak Time', x_label='Time (ns)', y_label='Counts')
res1 = plot_dist_tcuts(rdf, c1, pads1, ecal_cuts, 'time', hrange=(64, 0, 256), hthres=ecal_height_cut)
c1.SaveAs(os.path.join(args.out_dir, 'tcuts_ecal_time.png'))

# plot height distributions with t-cuts
c2 = ROOT.TCanvas('ECal_height', '', 1920, 1080)
pads2 = prepare_canvas(c2, ecal_channels, pad_label='Calorimeter Peak Height', x_label='ADC Value', y_label='Counts')
res2 = plot_dist_tcuts(rdf, c2, pads2, ecal_cuts, 'height', hrange=(100, 0, 500), hthres=ecal_height_cut)
c2.SaveAs(os.path.join(args.out_dir, 'tcuts_ecal_height.png'))


# join all the ecal channels
ecal_vec = '{' + ','.join(['{}_peak'.format(ch) for ch in ecal_cuts]) + '}'
ecal_vec_str = '{\"' + '\",\"'.join([ch for ch in ecal_cuts]) + '\"}'
# get the trigger peak and channel
rdf = rdf.Define('ecal', 'get_maximum_peak({})'.format(ecal_vec))\
         .Define('ecal_ch', 'std::vector<TString> ch_list{}; return ch_list[ecal.pos];'.format(ecal_vec_str))\
         .Filter('ecal.height > {}'.format(ecal_height_cut), 'Ecal Signal Cut')
save_list += ['ecal', 'ecal_ch']



"""
# =============================================================================
# Scint channel timing study
# =============================================================================
print('Study timing cuts for Scint2 channels...')
s2_channels = channels_layout['Scintillator2']
if args.read:
    s2_cuts = {c: json_data[c] for chs in s2_channels for c in chs}
else:
    s2_search = {c: {'cut': (-80, -40), 'ref': 'ecal.time'} for chs in s2_channels for c in chs}
    s2_cuts = timing_cuts_search(rdf, s2_search, cut_width=10)
    json_data.update(s2_cuts)

# apply cuts
for ch, cut in s2_cuts.items():
    rdf = rdf.Define('{}_peak'.format(ch), 'get_maximum_peak({}.peaks, time_cut_rel, {}, {}, {})'\
                                           .format(ch, *cut['cut'], cut['ref']))
s2_height_cut = 0

# plot distributions
c3 = ROOT.TCanvas('Scint2_time', '', 1920, 1080)
pads3 = prepare_canvas(c3, s2_channels, pad_label='Scintillator2 Peak Time',
                       x_label='Time Rel. to Ecal. (ns)', y_label='Counts')
res3 = plot_dist_tcuts(rdf, c3, pads3, s2_cuts, 'time', hrange=(80, -160, 160), hthres=s2_height_cut, use_ref=True)
c3.SaveAs(os.path.join(args.out_dir, 'tcuts_scint2_time.png'))

c4 = ROOT.TCanvas('Scint2_height', '', 1920, 1080)
pads4 = prepare_canvas(c4, s2_channels, pad_label='Scintillator2 Peak Height',
                       x_label='ADC Value', y_label='Counts')
res4 = plot_dist_tcuts(rdf, c4, pads4, s2_cuts, 'height', hrange=(100, 0, 500), hthres=s2_height_cut)
c4.SaveAs(os.path.join(args.out_dir, 'tcuts_scint2_height.png'))


# group all the scint2 channels
s2_vec = '{' + ','.join(['{}_peak'.format(ch) for ch in s2_cuts]) + '}'
s2_vec_str = '{\"' + '\",\"'.join([ch for ch in s2_cuts]) + '\"}'
# get the trigger peak and channel
rdf = rdf.Define('scint2', 'get_maximum_peak({})'.format(s2_vec))\
         .Define('scint2_ch', 'std::vector<TString> ch_list{}; return ch_list[scint2.pos];'.format(s2_vec_str))\
         .Filter('scint2.height > {}'.format(s2_height_cut), 'Scint2 Signal Cut')
save_list += ['scint2', 'scint2_ch']
"""



# =============================================================================
# Ending
# =============================================================================
rdf.Report().Print()
# save json
if not args.read:
    with open(args.json_file, 'w') as f:
        json.dump(json_data, f, indent=2)

# save data
if args.snap:
    print('Saving data snapshot to \"{}\", with columns:\n{}'.format(args.snap, save_list))
    rdf.Snapshot('EvTree', args.snap, save_list)

# this keeps the canvas
if not args.batch:
    print("press enter to quit")
    input()

